/*
*    INTEL CORPORATION PROPRIETARY INFORMATION 
*
*    This software is supplied under the terms of a license
*    agreement or nondisclosure agreement with Intel Corporation 
*    and may not be copied or disclosed except in accordance with 
*    the terms of that agreement.
*    Copyright (c) 1991,1992,1993  Intel Corporation. 
*
*/

/*++
	Win16Tst.C: WinMain() entry point

	Copyright Intel Corp. 1994

Description:
 	
	This program tests the new IA MM Extension instructions running in 
	16bit Protected mode environment (Win31)

	This module provides the Windows interface to display a message, and
	calls the test routine.

History:
-------
Nov-06-94- mikeh	Original

--*/
#include <windows.h>
#include <string.h>
#include <toolhelp.h>
#include "win16tst.h"
#include <afxres.h>
// Defines --------------------------------------
#define UNREFRENCED_PARAMS(x)	x;
#define IN
#define OUT

// Globals --------------------------------------
HWND ghDlg=0;
 
// Protos ---------------------------------------
LRESULT CALLBACK DlgProc(IN HWND hDlg, IN UINT message, IN WPARAM uParam, IN LPARAM lParam);
void PrintToEditor(IN char *pszText);
int testbase();
void emu_main(IN int argc, IN char *argv[]);
void RunIAMMxTest1(void);
void RunVxdTest(void);
void RunLoopTest1(DWORD dwCount);
void RunLoopTest2(DWORD dwCount);
void DoLoopTest(int iTestNumber);
void UseFpRegisters (IN BOOL fUseRegister);

// Functions ------------------------------------
              
              
int PASCAL WinMain(
	IN HINSTANCE hInstance,
    IN HINSTANCE hPrevInstance,
    IN LPSTR lpCmdLine,
    IN int nCmdShow)
/*++
Description:
	Entry point of program. Displays the Dialog box "IDD_DIALOG".

Return:
	always 0
--*/
{
	DialogBox(hInstance, "IDD_DIALOG", NULL, (DLGPROC)DlgProc);

	return (0);
   	UNREFRENCED_PARAMS((lpCmdLine,nCmdShow,hPrevInstance));
}

LRESULT CALLBACK DlgProc(
	IN HWND hDlg,           // window handle of the dialog box
    IN UINT message,        // type of message
    IN WPARAM uParam,       // message-specific information
    IN LPARAM lParam) 		
/*++
Description:
	Dialog Proc handler.  Has one list box that acts like a scrollable 
	output text buffer.  The dialog proc runs the testbase() which is
	a set of mmx instructios tests.

Return:
	TRUE: if it handled the message,
	FALSE: otherwise..
--*/

{
	ghDlg = hDlg;
	
	switch(message) {
	case WM_INITDIALOG:
		return TRUE;

   	case WM_COMMAND:                      
		
		switch (LOWORD(uParam)) {
    	case IDCANCEL: 				// System menu close command?
       		EndDialog(hDlg, TRUE);  // Exit the dialog
			return TRUE;
 
		case ID_16BITPMTEST:
			SendMessage(GetDlgItem(ghDlg, IDC_LIST1), LB_RESETCONTENT, 0, 0);	
		    RunIAMMxTest1();
			return TRUE;
					
		case ID_32BITFLATVXDTEST:
			SendMessage(GetDlgItem(ghDlg, IDC_LIST1), LB_RESETCONTENT, 0, 0);	
			RunVxdTest();
			return TRUE;
		
		case ID_TIMINGTESTS_EMMS:
			SendMessage(GetDlgItem(ghDlg, IDC_LIST1), LB_RESETCONTENT, 0, 0);	
			UseFpRegisters(TRUE);
			PrintToEditor("Using FP registers.. Wait Test in progress..");
			DoLoopTest(1);
			UseFpRegisters(FALSE);
			PrintToEditor("Not Using FP registers.. Wait Test in progress..");
			DoLoopTest(1);
			return TRUE;

		case ID_TIMINGTESTS_MOVQANDPMADDWD:
			SendMessage(GetDlgItem(ghDlg, IDC_LIST1), LB_RESETCONTENT, 0, 0);	
			UseFpRegisters(TRUE);
			PrintToEditor("Using FP registers.. Wait Test in progress..");
			DoLoopTest(2);
			UseFpRegisters(FALSE);
			PrintToEditor("Not Using FP registers.. Wait Test in progress..");
			DoLoopTest(2);
			return TRUE;

		case ID_APP_ABOUT:
			MessageBox (hDlg, "Win16Tst.Exe - Ver 1.00.005 - Aug 15, 1995", 
				"About", MB_OK|MB_ICONINFORMATION);
			return TRUE;
		}
	}

    return (FALSE);

	UNREFRENCED_PARAMS(lParam); 
}
    

void
DoLoopTest(int iTestNumber)
{
	TIMERINFO startInfo;
	TIMERINFO endInfo;
	char szTmp[256];
	DWORD dwCount=0x100000;
	
	startInfo.dwSize = sizeof(TIMERINFO);
	endInfo.dwSize = sizeof(TIMERINFO);
	TimerCount(&startInfo);
	
	switch (iTestNumber) {
		case 1:	RunLoopTest1(dwCount);	break;
		case 2:	RunLoopTest2(dwCount);  break;
	}		              
	
	TimerCount(&endInfo);
	wsprintf (szTmp, "ElapsedTime: %08lx, Count: %08lx",
		endInfo.dwmsThisVM - startInfo.dwmsThisVM, dwCount);
	PrintToEditor(szTmp);
}

void
UseFpRegisters (
	IN BOOL fUseRegister
	)
{
	DWORD dwVxdEntry;
	
	_asm {          
		mov		ax, 1684h
		mov		bx, 5a5ah
		int		2fh
		mov		WORD PTR dwVxdEntry, di
		mov		WORD PTR dwVxdEntry[2], es
	}

	if (dwVxdEntry == 0) {
		MessageBox (ghDlg, "VmmxEmu.386 Not detected.. "
			"Refer to installation documentation for proper installation procedure", "ERROR", MB_OK);
		return;
		}

	// Now call the VxD to perform the test..
	if (fUseRegister)
		_asm	mov	dx, 0
	else
		_asm	mov	dx, 1
			
	_asm call	DWORD PTR dwVxdEntry
	
}    

void
CompareResults (
	char far *pszInst, 
	unsigned long ulResultHigh,
	unsigned long ulResultLow,
	unsigned long ulExpectedHigh,
	unsigned long ulExpectedLow
	)
{
	char szTmp[256];
	
	if ( (ulResultHigh == ulExpectedHigh) && (ulResultLow == ulExpectedLow) ) {
		wsprintf (szTmp, "%s PASSED: Expected Result: 0x%08lx 0x%08lx: got  0x%08lx 0x%08lx",
			pszInst, ulExpectedHigh, ulExpectedLow, ulResultHigh, ulResultLow);
		PrintToEditor(szTmp);
	} else {
		wsprintf (szTmp, "***********%s FAILED: Expected Result: 0x%08lx 0x%08lx: got  0x%08lx 0x%08lx",
			pszInst, ulExpectedHigh, ulExpectedLow, ulResultHigh, ulResultLow);
		PrintToEditor(szTmp);
	}		

}

void
RunVxdTest(void)
{

	DWORD dwVxdEntry;
	const char *szMessage1= "Hey, why don't I see anything here.. Here is the story.. ";
	const char *szMessage2= "In this mode Win16Tst.Exe calls VmmxTst.386 to start a ";
	const char *szMessage3= "Ring0 32bit Flat mode test.  Since VxDs don't have an easy ";
	const char *szMessage4= "access to Windows APIs, the results are sent to the debug port. ";
	const char *szMessage5= "So if you run WinIce and then run this test, you'll see the output ";
	const char *szMessage6= "On WinIce..";

	_asm {          
		mov		ax, 1684h
		mov		bx, 1234h
		int		2fh
		mov		WORD PTR dwVxdEntry, di
		mov		WORD PTR dwVxdEntry[2], es
	}

	if (dwVxdEntry == 0) {
		MessageBox (ghDlg, "VmmxTst.386 Not detected.. "
			"Refer to installation documentation for proper installation procedure", "ERROR", MB_OK);
		return;
		}

	PrintToEditor ((char *)szMessage1);
    PrintToEditor ((char *)szMessage2);
	PrintToEditor ((char *)szMessage3);
	PrintToEditor ((char *)szMessage4);
	PrintToEditor ((char *)szMessage5);
	PrintToEditor ((char *)szMessage6);

	// Now call the VxD to perform the test..
	_asm {
		mov		dx, 0
		call	DWORD PTR dwVxdEntry
		}
}

void PrintToEditor(
	IN char *pszText
	)
/*++
Description:
	Prints the string pszText to the output buffer. (In our case 
	it is a list box.)

Return:
	who cares

--*/
{
	static int index=0;
	HWND hListBox=GetDlgItem(ghDlg, IDC_LIST1);

	// add the string to the list box,and set hightlight to it
	// so it would get displayed.
	SendMessage(hListBox, LB_ADDSTRING, 0, (LPARAM)pszText);

	if (index<30)
		SendMessage(hListBox, LB_SETCURSEL, index++, 0);
}

