/*
*    INTEL CORPORATION PROPRIETARY INFORMATION 
*
*    This software is supplied under the terms of a license
*    agreement or nondisclosure agreement with Intel Corporation 
*    and may not be copied or disclosed except in accordance with 
*    the terms of that agreement.
*    Copyright (c) 1991,1992,1993  Intel Corporation. 
*
*/
     

/*++
	Win32Tst.C: WinMain() entry point

	Copyright Intel Corp. 1994

Description:
 	
	This program tests the new IA MM Extension instructions running in 
	32bit mode environment (NT, Win32S, Chicago).

	This module provides the Windows interface to display a message, and
	calls the test routine.

History:
-------
Aug-15-94- mikeh	Original

--*/
#include <windows.h>
#include <string.h>
#include <mmsystem.h>
#include "win32tst.h"
#include <afxres.h>

#include "mmuser.h"

// Defines --------------------------------------
#define UNREFRENCED_PARAMS(x)	x;
#define IN
#define OUT

// Globals --------------------------------------
HWND ghDlg=0;
static BOOL sFailed;


_declspec (dllimport) BYTE *pMmReg;

// Protos ---------------------------------------
LRESULT CALLBACK DlgProc(IN HWND hDlg, IN UINT message, IN WPARAM uParam, IN LPARAM lParam);
void PrintToEditor(IN CHAR *pszText);
void CExceptionEmulate();
void RunIAMMxTest1();
void decode();
void RunLoopTest1(DWORD dwCount);
void RunLoopTest2(DWORD dwCount);
void DoLoopTest(int iTestNumber);
BOOL ValidWinVersion ();


// Functions ------------------------------------

int APIENTRY WinMain(
	IN HINSTANCE hInstance,
    IN HINSTANCE hPrevInstance,
    IN LPSTR lpCmdLine,
    IN int nCmdShow)
/*++
Description:
	Entry point of program. Displays the Dialog box "IDD_DIALOG".

Return:
	always 0
--*/
{

	BYTE *pReg = pMmReg;

	DialogBox(hInstance, "IDD_DIALOG", NULL, DlgProc);

	return (0);
   	UNREFRENCED_PARAMS((lpCmdLine,nCmdShow,hPrevInstance));
}

LRESULT CALLBACK DlgProc(
	IN HWND hDlg,           // window handle of the dialog box
    IN UINT message,        // type of message
    IN WPARAM uParam,       // message-specific information
    IN LPARAM lParam) 		
/*++
Description:
	Dialog Proc handler.  Has one list box that acts like a scrollable 
	output text buffer.  The dialog proc runs the testbase() which is
	a set of mmx instructios tests.

Return:
	TRUE: if it handled the message,
	FALSE: otherwise..
--*/

{

	ghDlg = hDlg;

	switch(message) {
		case WM_INITDIALOG:
			return TRUE;
 
   		case WM_COMMAND:                      
			switch (LOWORD(uParam)) {
        		case IDCANCEL: 				// System menu close command?
            		EndDialog(hDlg, TRUE);  // Exit the dialog
					return TRUE;

				case ID_EMULATIONUSINGVXD:
					if (!ValidWinVersion())
						break;

        			SendMessage(GetDlgItem(ghDlg, IDC_LIST1), LB_RESETCONTENT, 0, 0);
					RunIAMMxTest1();
					break;

				case ID_EMULATIONUSINGCEXCEPTIONS:
        			SendMessage(GetDlgItem(ghDlg, IDC_LIST1), LB_RESETCONTENT, 0, 0);
					CExceptionEmulate(1);
					break;

				case ID_TIMINGTESTS_EMMS:
					DoLoopTest(1);
					return TRUE;

				case ID_TIMINGTESTS_MOVQANDPMADDWD:
					DoLoopTest(2);
					return TRUE;

				case ID_TIMINGTESTS_CEXCEPTIONLOOP:
        			SendMessage(GetDlgItem(ghDlg, IDC_LIST1), LB_RESETCONTENT, 0, 0);
					CExceptionEmulate(100);
					return TRUE;

				case ID_INVALIDPAGEACCESS_ACCESSINVALIDSEGMENT:
					AccessInvalidSegment();
					return TRUE;

				case ID_INVALIDPAGEACCESS_ACCESSINVALIDADDRESS:
					AccessInvalidData();
					return TRUE;

        		case ID_APP_ABOUT:
		        	MessageBox (hDlg, "Win32Tst.Exe - Ver 1.00.005 - Aug 15, 1995", 
				        "About", MB_OK|MB_ICONINFORMATION);
			        return TRUE;
				}

           	break;
		}

    return (FALSE); 	// Didn't process the message

	UNREFRENCED_PARAMS(lParam); 
}

void
DoLoopTest(int iTestNumber)
{
	DWORD startTime;
	DWORD endTime;
	char szTmp[256];
	DWORD dwCount=0x100000;
	
//	if (!ValidWinVersion())
//		return;

	SendMessage(GetDlgItem(ghDlg, IDC_LIST1), LB_RESETCONTENT, 0, 0);
    PrintToEditor(".. Wait Test in progress...");
	startTime = timeGetTime();	
	_try {
	switch (iTestNumber) {
		case 1:	RunLoopTest1(dwCount);	break;
		case 2:	RunLoopTest2(dwCount);  break;
	}
	}
	_except (MMX_HANDLE_EXCEPTION)	{}

	endTime = timeGetTime();
	wsprintf (szTmp, "ElapsedTime: %08lx, Count: %08lx",
		endTime - startTime, dwCount);
	PrintToEditor(szTmp);
}


void PrintToEditor(
	IN CHAR *pszText 
	)
/*++
Description:
	Prints the string pszText to the output buffer. (In our case 
	it is a list box.)

Return:
	who cares

--*/
{
	static int index=0;
	HWND hListBox=GetDlgItem(ghDlg, IDC_LIST1);

	// add the string to the list box,and set hightlight to it
	// so it would get displayed.
	SendMessage(hListBox, LB_ADDSTRING, 0, (LPARAM)pszText);

	if (index<30)
		SendMessage(hListBox, LB_SETCURSEL, index++, 0);
}

void
CompareResults (
	char far *pszInst, 
	unsigned long ulResultHigh,
	unsigned long ulResultLow,
	unsigned long ulExpectedHigh,
	unsigned long ulExpectedLow
	)
{
	char szTmp[256];

	if ( (ulResultHigh == ulExpectedHigh) && (ulResultLow == ulExpectedLow) ) {
		wsprintf (szTmp, "%s PASSED: Expected Result: 0x%08lx 0x%08lx: got  0x%08lx 0x%08lx",
			pszInst, ulExpectedHigh, ulExpectedLow, ulResultHigh, ulResultLow);
		PrintToEditor(szTmp);
	} else {
        sFailed++;
		wsprintf (szTmp, "********** %s FAILED: Expected Result: 0x%08lx 0x%08lx: got  0x%08lx 0x%08lx",
			pszInst, ulExpectedHigh, ulExpectedLow, ulResultHigh, ulResultLow);
		PrintToEditor(szTmp);
	}		
}

void	
CExceptionEmulate(int loopCount)
/*++
Description:
	This is the _try/_except mode to emulate IAMMx instructions. All IAMMx instructions
	should be inside the _try block.. This goes also for any function that uses IAMMx
	instructions..

	MMX_HANDLER_EXCEPTION expands to a call to MMxHandleException() with the context info passed 
	to it as parameters.  MMxHandleException() lives in mm.c

Return:
	none that I know of.

--*/
{
    char szTmp[50];
	int i=0;

    sFailed = 0;

	_try {
		for (i=0;i<loopCount; i++)
			RunIAMMxTest1();
	}
	_except (MMX_HANDLE_EXCEPTION)	{
		/*** this point is never reached ***/
	}

    if (sFailed) {
        wsprintf (szTmp, "%d instruction(s) failed\n", sFailed);
        MessageBox(NULL, szTmp, "ERROR", MB_OK);
    }
}

BOOL
ValidWinVersion ()
/*++
Description:
	Checks to make sure that the app is not running on Windows NT when the 
	VxD mode is required..

Return:
	TRUE: if not Windows NT.
	FLASE, O.W.

--*/

{
	const char *szInvOsMessage = 
		"Aaah, you must be using Windows NT.. "
		"Unfortunatly, this option is only available under "
		"Win31 w/Win32s installed and Win95 (Chicago). "
		"Follow the guidlines in the release notes to install it "
		"and run it.";
	const char *szInvOsTitle = "Noo Noo Noo";

	DWORD dwVersion;


 	if (dwVersion = GetVersion()) {
 		if ( (dwVersion & 0x80000000) == 0) {
 			MessageBox (ghDlg, szInvOsMessage, szInvOsTitle, MB_OK);
 			return FALSE;
 		}
 	}
	
	return TRUE;
}
