/*
*    INTEL CORPORATION PROPRIETARY INFORMATION 
*
*    This software is supplied under the terms of a license
*    agreement or nondisclosure agreement with Intel Corporation 
*    and may not be copied or disclosed except in accordance with 
*    the terms of that agreement.
*    Copyright (c) 1991,1992,1993  Intel Corporation. 
*
*/

/*

   How should mm package need to be used ?
   ---------------------------------------------
   If Unix app should call start_mmx before executing the first mmx code
   Elseif Window this module need to be compiled with WINNT compilation switch 
     If Console application (I.e. running in DOS prompt)
       user code should resides in real_main(int argc, char *argv[]).
     If Window Application then this module need to be compiled with WIN_APP compilation 
       switch. The Window application should call emu_main routine before first mmx
       instruction. This routine will call real_main in the user code. All mmx
       instructions should be used only in this routine trees. 
       User Code should supply a printing function PrintToEditor and a line buffer 
       lineBuf.
   
   What about mmx code in the driver ?
   ------------------------------------
   Currently, mmx instructions executed by driver will not be
   simulated. For this, this code should be turned into a
   driver that will serve the whole system (including non-mmx
   applications). At this stage we know how to grab all interrupt 
   services, before we'll make this change we need to find out
   how to cascade it back to the system Interrupt handler. This
   will take care of handling other interrupts, as well as doing
   the right thing when a non-mmx instruction is encountered.

   What is not being simulated ?
   -----------------------------
   Aliasing the mm register in the fp registers (instead they are kept in the user
   application's memory space), executing EMMS, treating prefixes.
   As far as we know only Segment Override is important (in Windows application).
   
*/
#include <stdio.h>
#include <stdlib.h>
#include "mm.h"
#include "mmall.h"
#include "mmx.h"


#define FAST_EMULATE
#ifdef FAST_EMULATE
	#ifdef WIN_VXD
		#define MAX_COUNT 500
	#endif
	#ifdef WIN_APP
		#define MAX_COUNT 10000
	#endif
	#ifndef MAX_COUNT  /* unix not tested */
		#define MAX_COUNT 1
	#endif
#else
	#define MAX_COUNT 1
#endif	


#ifndef CONTEXT_i486
typedef struct _CONTEXT {
	FP_CONTEXT	FloatSave;
}CONTEXT;
#endif

CONTEXT		Context;

#if defined(WIN_VXD)

#define MCONTEXT_FLOATING_POINT		0x000010008

#include <vmm.h>
#include "myvkd.h"
extern DWORD ghThread;
void rotate_context2( FLOATING_SAVE_AREA *pCont );
int _cdecl Get_Thread_Context(DWORD hThread, PVOID pContext);
int _cdecl Set_Thread_Context(DWORD hThread, PVOID pContext);
BOOL CheckAddressRange(DWORD dwAddress, DWORD nBytes);

		#define SAVE_FP_STATE() 	\
			Context.ContextFlags = MCONTEXT_FLOATING_POINT; \
			Get_Thread_Context(ghThread, &Context); 	\
			rotate_context2( &Context.FloatSave );
		#define RESTORE_FP_STATE()  \
			Context.ContextFlags = MCONTEXT_FLOATING_POINT; \
			Set_Thread_Context(ghThread, &Context);
#else
	#define SAVE_FP_STATE()
	#define RESTORE_FP_STATE()
#endif	



/************** GLOBALS ***********************************************/
DWORD  instr_size;
DWORD  gdwDsSegSelector;
DWORD  gdwSsSegSelector;
BOOL   gfStackSegUsed;
DWORD  gdwUseFpRegisters=TRUE;
DWORD  gSize;
DWORD  gLastInstrSize;
DWORD  gBytesProcessed;
BOOL   gfMmRegUsed;
DWORD  gfSingleStepEnabled=FALSE;
BYTE   *gpCurInstrStartEip;

FP_CONTEXT  *gFpContext = (FP_CONTEXT*) &Context.FloatSave;
U80 *pMmReg = (U80 *)&Context.FloatSave.RegisterArea;

static BYTE ValidMmxOpcode[] =
/*		0	1	2	3	4	5	6	7	8	9	a	b	c	d	e	f */
	{	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	/* 0 */
		0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	/* 1 */
	   	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	/* 2 */
		0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	/* 3 */
		0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	/* 4 */
		0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	/* 5 */
		1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	1,	0,	0,	1,	1,	/* 6 */
		0,	1,	1,	1,	1,	1,	1,	1,	0,	0,	0,	0,	0,	0,	1,	1,	/* 7 */
		0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	/* 8 */
		0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	/* 9 */
		0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	/* a */
		0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	/* b */
		0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	0,	/* c */
		0,	1,	1,	1,	0,	1,	0,	0,	1,	1,	0,	1,	1,	1,	0,	1,	/* d */
		0,	1,	1,	0,	0,	1,	0,	0,	1,	1,	0,	1,	1,	1,	0,	1,	/* e */
		0,	1,	1,	1,	0,	1,	0,	0,	1,	1,	1,	0,	1,	1,	1,	0,	/* f */
	};


#ifdef WIN_VXD
#define ERROR_MSG 		
	DWORD gdwLastEipReg;
	DWORD gdwLastCsReg;
	BYTE  gLastOpcode[20];
	WORD  gLastOpcodeSize=0;

	char ERR_GENERAL[] = 	"Hit:\n"
		".   [Enter] to try to skip this instruction and continue execution of\n"
		".           the application\n"  										 
		".   [Esc]   to insert an 'INT 3' break point instruction at the\n"		 
		".           faulting address.  This allows the debugger to break\n"	 
		".           at that address.\n"										 
		".           NOTE: that here, the first byte of the instruction \n" 	 
		".           is replaced with 0xcc(int3), and the debugger doesn't\n"	 
		".           know about it.  YOU MUST EXIT AND RESTART YOUR APPLICATION\n"
		".           OTHERWISE, YOU'LL GET A GP FAULT.";

	char ERR_INVALID_SEGMENT[] = 
		"Invalid Data or Stack Segment specified\n\n";

	char ERR_FATAL_ERROR_OCCURED[] = 
		"Some FATAL ERROR HAPPNED\n\n";


	char ERR_INVALID_SEGMENT_LIMIT[] = 
		"Invalid Segment Limit\n\n";

	char ERR_SEGMENT_MAPPING_FAILED[] = 
	 	"Segment mapping failed\n\n";


	char ERR_PAGENOTPRESENT[] = 
		"Page Fault: Linear Address Page not present.\n\n";

	char ERR_RDPMC_INVALID_PARM[] = 
		"RDPMC: specified index in ECX is invalid\n\n";
#endif


/************** Prototypes ***********************************************/
void	fatal(char *msg, int sig);
static void movd_st ( MMX64 *dst, MMX64 *src , int x86reg ,int dst_index);
static void movd_ld ( MMX64 *src, MMX64 *dst , int x86reg ,int src_index);
static void movq ( MMX64 *src, MMX64 *dst);	
static void rdpmc();
#ifndef UNIX_APP
DisplayErrorBox (char *szMsg);
ReturnFaultToDebugger();
#endif




/*-------------------------------------------------------------------------*/
static int 
get_d8() 
{
	int val;
#ifdef WIN_VXD
	CheckAddressRange(R__EIP, 1);
#endif
	val = *(((char *) R__EIP) + gSize);
	gSize++;
	return val;
}

/*-------------------------------------------------------------------------*/
static int
get_d16() 
{	
	int val;
#ifdef WIN_VXD
	CheckAddressRange(R__EIP, 2);
#endif
	val = *((short *) (((char *) R__EIP) + gSize));
	gSize += 2;
	return val;
}

/*-------------------------------------------------------------------------*/
static int
get_d32() 
{	
	int val;
#ifdef WIN_VXD
	CheckAddressRange(R__EIP, 4);
#endif	
	val = *((int *) (((char *) R__EIP) + gSize));
	gSize += 4;
	return val;
}


/*-------------------------------------------------------------------------*/
static unsigned long 
get_sib(mod)  int mod;
{
	unsigned char  sib = *(((char *) R__EIP) + 3);
	int scale = sib >> 6;
	int base = sib & 7;
	int index = (sib >> 3) & 7;
	int s_index;
	switch (index)
	{
		case 0: s_index = R__EAX << scale; break;
		case 1: s_index = R__ECX << scale; break;
		case 2: s_index = R__EDX << scale; break;
		case 3: s_index = R__EBX << scale; break;
		case 4: s_index = 0;               break;
		case 5: s_index = R__EBP << scale; break;
		case 6: s_index = R__ESI << scale; break;
		case 7: s_index = R__EDI << scale; break;
	}
	gSize ++;
	switch (base)
	{
		case 0: return R__EAX + s_index;
		case 1: return R__ECX + s_index;
		case 2: return R__EDX + s_index;
		case 3: return R__EBX + s_index;
		case 4: 
			gfStackSegUsed=TRUE;
			return R__ESP + s_index;
		case 5: 
			if ( (mod == 0x40) || (mod == 0x80) ) {
				gfStackSegUsed=TRUE;
				 return R__EBP + s_index;
			}
			else
				 return get_d32() + s_index;
		case 6: return R__ESI + s_index;
		case 7: return R__EDI + s_index;
	}
}


/*-------------------------------------------------------------------------*/

void rotate_context()
{	U80		fp_temp[8];			/* 8 FP regs, 80 bits each. */
	int tos = ((gFpContext->StatusWord >> 11 ) & 7);
	int i;
	
	for ( i = 0 ; i < 8 ; i++)
		fp_temp[i] = 	gFpContext->FpRegisters[(i-tos) & 7 ];

	for ( i = 0 ; i < 8 ; i++)
		gFpContext->FpRegisters[i] = fp_temp[i];
}
/* Here, just decode up to MAX_COUNT consecutive instructions.
 * Return to app on the first non MMx instruction, or a fault
 * happened.
 */
void fastDecode () 
{
	int count = 0;

	gBytesProcessed = 0;
	gLastInstrSize = 0;

	do {
		gpCurInstrStartEip = (BYTE*)R__EIP;
		if (!decode()) {
			/* for VxD & UNIX, do fatal() */
			gSize = 0;
			return;
		}


		/* increment EIP to next instruction, and increment global
		 * bytes processed so far..
		 */
		gLastInstrSize = gSize;
		R__EIP += gSize;
		gBytesProcessed += gSize;

		/* if the next instruction deosn't start with 0F go home.. */

		if ( ((unsigned char *) R__EIP)[0] != 0x0F )  {
			break;
		}

	} while (!gfSingleStepEnabled && (count++ < MAX_COUNT) );
}
											 
BOOL decode() {
	unsigned char  modrm,opcode, OpCodeByte2;
	unsigned char wMod, wReg, wRm;
	int source;
	unsigned int addr;
	MMX64 *src_ptr;
	BYTE bFirstByte =  ((BYTE *) R__EIP)[0];
	BYTE bSecondByte = ((BYTE *) R__EIP)[1];
	static DWORD DebugCounter=0;

	DebugCounter++;

	/* 0x0f for possible mmx opcode, 0xcc(int3) to overcome msvc20 debugger
	 * problem, where it puts 0xcc instead of 0x0f when you do a STEP with
	 * the debugger.  It put 0xcc because it tries to decode the mmx
	 * instruction to put the int3 after it; however since it doesn't 
	 * understand mmx, it puts the int3 at the current IP..
	 */
	
	if ( (bFirstByte!= 0x0F && bFirstByte != 0xcc) || !ValidMmxOpcode[bSecondByte] )  {
		return FALSE;
	}

    gfMmRegUsed=FALSE;
	gfStackSegUsed=FALSE;			/* assumed DS segment */
	gSize = 3;

	/* here we try to use the values of the floating point registers
	 * assuming they were saved by the OS after a context switch..
	 */
	if (gdwUseFpRegisters) {
		SAVE_FP_STATE();
		gFpContext->StatusWord &= 0xffffc7ff;  	/* mask off TOP (bit 11,12,13) */
		gFpContext->TagWord    &= 0xffff0000;		/* clear tag word (see emms below). */
	 }

#ifdef WIN_VXD	
	CheckAddressRange(R__EIP, 3);
#endif
	opcode = *(((char *) R__EIP) +1);
	OpCodeByte2 = *(((char *) R__EIP) +2);
	source = (OpCodeByte2 >> 3) & 0x7;
	wMod  = OpCodeByte2 & 0xc0;
	wReg  = OpCodeByte2 & 0x38;
	wRm	  = OpCodeByte2 & 0x07;
	modrm = wMod | wRm;

	/* Fast response to these instructions */
	switch (opcode) {
#ifdef WIN_VXD
		case Rdpmc:
			gSize =2;
			rdpmc();
			goto Heaven;
#endif
		case Emms:
			gFpContext->TagWord |= 0xffff;		/* set tag word (invalid values). */
			gSize=2;
			goto Heaven;
	}
	
#ifdef WIN_VXD
	if (gf32BitExecMode) {
#endif
		/* Process the R/M bits now */
		switch (wRm) {
		case 0x0: addr = R__EAX; break;		
		case 0x1: addr = R__ECX; break;
		case 0x2: addr = R__EDX; break;
		case 0x3: addr = R__EBX; break;
		case 0x4: if (wMod < 0xc0) addr = get_sib(wMod); break;
		case 0x5: 
			if (wMod==0)
				addr = get_d32();
			else if (wMod < 0xc0) {
				addr = R__EBP;
				gfStackSegUsed = TRUE;
				}
			break;
		case 0x6: addr = R__ESI; break;
		case 0x7: addr = R__EDI; break;
		}

		/* Process the MOD bits now */
		switch(wMod) {
		case 0x00: break;
		case 0x40: addr += get_d8(); break;
		case 0x80: addr += get_d32(); break;
		case 0xc0: addr = (unsigned int) &pMmReg[wRm]; gfMmRegUsed=TRUE; break; 
		}

#ifdef WIN_VXD
	}else{	/* 16 bit apps (PM and V86). */
		/* Process the R/M bits first */
		switch (wRm) {
		case 0x0: addr = R__EBX + R__ESI; break;
		case 0x1: addr = R__EBX + R__EDI; break;
		case 0x2: addr = R__EBP + R__ESI; gfStackSegUsed = TRUE; break;
		case 0x3: addr = R__EBP + R__EDI; gfStackSegUsed = TRUE; break;
		case 0x4: addr = R__ESI; break;
		case 0x5: addr = R__EDI; break;
		case 0x6:
			if (wMod == 0)
				addr = get_d16();
			else if (wMod < 0xc0) {
				addr =  R__EBP;
				gfStackSegUsed = TRUE;
				}
			break;
		case 0x7: addr = R__EBX; break;
		}
		addr &= 0xffff;		/* Only need low 16 bits in address */

		/* Process the MOD bits now */
		switch(wMod) {
		case 0x40: addr += get_d8(); break;
		case 0x80: addr += get_d16(); break;
		case 0xc0: addr = (unsigned int ) &pMmReg[wRm]; gfMmRegUsed=TRUE; break; 
		}
	}

    /* for access to non-MM register only */
	if ( (wMod != 0xc0) ) {
		gdwUsedSegSelector = (gfStackSegUsed) ? gdwSsSegSelector: gdwDsSegSelector;

		if ( !IsSegmentValid(gdwUsedSegSelector) )
			ReturnFaultToDebugger(ERR_INVALID_SEGMENT);
	}

#endif

	if ( (source < 0) || (source > 7) )
		_asm int 3;

	src_ptr = (MMX64 *)&pMmReg[source];

	if (opcode == PSHimd || opcode ==  PSHimq || opcode ==  PSHimw)
		gFpContext->FpRegisters[wRm].Exponent = (WORD)-1;
	else if (opcode != Movd_st && opcode != Movq_st) 
		gFpContext->FpRegisters[source].Exponent = (WORD)-1;

#ifdef WIN_VXD
	/* Make sure that the memory to be accessed is accessible. */
	switch (opcode) {
        case Movd_ld: 		/* these two are done in their handlers.. */
        case Movd_st: 
			break;
			 
		case Movq_ld: 
			if (!MapAddressFlat( &((MMX64*)addr), 8, MODE_READACCESS))
				return FALSE;
			break;
		case Movq_st: 
			if (!MapAddressFlat( &((MMX64*)addr), 8, MODE_WRITEACCESS))
				return FALSE;
			break;
        case Punpcklbw: 
        case Punpckldq: 
        case Punpcklwd: 
        	if (!MapAddressFlat( &((MMX64*)addr), 4, MODE_READACCESS))
				return FALSE;
			break;

		default:
			if (!MapAddressFlat( &((MMX64*)addr), 8, MODE_READACCESS))
				return FALSE;
	}
#endif

	/* Emms and rdpmc are handled eariler */
	switch(opcode) {
        case Movd_ld: movd_ld((MMX64 *) addr,src_ptr,
		                 wMod == 0xC0 ,wRm); break;
        case Movd_st: movd_st((MMX64 *) addr,src_ptr,
		                 wMod == 0xC0 ,wRm); break;
        case Movq_ld: movq((MMX64 *) addr,src_ptr); break;
        case Movq_st: movq(src_ptr,(MMX64 *) addr); break;
        case Packsswb: packsswb(src_ptr,(MMX64 *) addr); break;
        case Packssdw: packssdw(src_ptr,(MMX64 *) addr); break;
        case Packuswb: packuswb(src_ptr,(MMX64 *) addr); break;
        case Paddb: paddb(src_ptr,(MMX64 *) addr); break;
        case Paddd: paddd(src_ptr,(MMX64 *) addr); break;
        case Paddsb: paddsb(src_ptr,(MMX64 *) addr); break;
        case Paddsw: paddsw(src_ptr,(MMX64 *) addr); break;
        case Paddusb: paddusb(src_ptr,(MMX64 *) addr); break;
        case Paddusw: paddusw(src_ptr,(MMX64 *) addr); break;
        case Paddw: paddw(src_ptr,(MMX64 *) addr); break;
        case Pand: pand(src_ptr,(MMX64 *) addr); break;
        case Pandn: pandn(src_ptr,(MMX64 *) addr); break;
        case Pcmpeqb: pcmpeqb(src_ptr,(MMX64 *) addr); break;
        case Pcmpeqd: pcmpeqd(src_ptr,(MMX64 *) addr); break;
        case Pcmpeqw: pcmpeqw(src_ptr,(MMX64 *) addr); break;
        case Pcmpgtb: pcmpgtb(src_ptr,(MMX64 *) addr); break;
        case Pcmpgtd: pcmpgtd(src_ptr,(MMX64 *) addr); break;
        case Pcmpgtw: pcmpgtw(src_ptr,(MMX64 *) addr); break;
        case Pmaddwd: pmaddwd(src_ptr,(MMX64 *) addr); break;
        case Pmulhw: pmulhw(src_ptr,(MMX64 *) addr); break;
        case Pmullw: pmullw(src_ptr,(MMX64 *) addr); break;
        case Por: por(src_ptr,(MMX64 *) addr); break;
        case PSHimd: 
			switch ( wReg ) {
			case 0x30: pslld_imm8((MMX64 *) addr,(MMXU8) get_d8()); break;
			case 0x10: psrld_imm8((MMX64 *) addr,(MMXU8) get_d8()); break; 
			case 0x20: psrad_imm8((MMX64 *) addr,(MMXU8) get_d8()); break;
			default:
				return FALSE;
			}
			break;
        case PSHimq: 
			switch ( wReg ) {
			case 0x30: psllq_imm8((MMX64 *) addr,(MMXU8) get_d8()); break;
			case 0x10: psrlq_imm8((MMX64 *) addr,(MMXU8) get_d8()); break; 
			default:
				return FALSE;
			}
			break;
		case PSHimw: 
			switch ( wReg ) {
			case 0x30: psllw_imm8((MMX64 *) addr,(MMXU8) get_d8()); break;
			case 0x10: psrlw_imm8((MMX64 *) addr,(MMXU8) get_d8()); break; 
			case 0x20: psraw_imm8((MMX64 *) addr,(MMXU8) get_d8()); break;
			default:
				return FALSE;
			}
			break;
        case Pslld: pslld(src_ptr,(MMX64 *) addr); break;
        case Psllq: psllq(src_ptr,(MMX64 *) addr); break;
        case Psllw: psllw(src_ptr,(MMX64 *) addr); break;
        case Psrad: psrad(src_ptr,(MMX64 *) addr); break;
        case Psraw: psraw(src_ptr,(MMX64 *) addr); break;
        case Psrld: psrld(src_ptr,(MMX64 *) addr); break;
        case Psrlq: psrlq(src_ptr,(MMX64 *) addr); break;
        case Psrlw: psrlw(src_ptr,(MMX64 *) addr); break;
        case Psubb: psubb(src_ptr,(MMX64 *) addr); break;
        case Psubsb: psubsb(src_ptr,(MMX64 *) addr); break;
        case Psubd: psubd(src_ptr,(MMX64 *) addr); break;
        case Psubsw: psubsw(src_ptr,(MMX64 *) addr); break;
        case Psubusb: psubusb(src_ptr,(MMX64 *) addr); break;
        case Psubusw: psubusw(src_ptr,(MMX64 *) addr); break;
        case Psubw: psubw(src_ptr,(MMX64 *) addr); break;
        case Punpcklbw: punpcklbw(src_ptr,(MMX64 *) addr); break;
        case Punpckldq: punpckldq(src_ptr,(MMX64 *) addr); break;
        case Punpcklwd: punpcklwd(src_ptr,(MMX64 *) addr); break;
        case Punpckhbw: punpckhbw(src_ptr,(MMX64 *) addr); break;
        case Punpckhdq: punpckhdq(src_ptr,(MMX64 *) addr); break;
        case Punpckhwd: punpckhwd(src_ptr,(MMX64 *) addr); break;
        case Pxor: pxor(src_ptr,(MMX64 *) addr); break;
		default:
			gSize = 0;
			return FALSE;
	}

Heaven:
	if (gdwUseFpRegisters) {
		RESTORE_FP_STATE();
	}
	return TRUE;
}

#ifdef WIN_VXD
static	void	fatal(char *msg, int sig)
{
	DisplayErrorBox (ERR_FATAL_ERROR_OCCURED);
}

BOOL
IsSegmentValid(DWORD dwSegmentSelector)
{
	DWORD dwPrvLevel;
	SegDescriptor_t Descriptor;

	if (!GetDescriptorFields (dwSegmentSelector, &Descriptor))
		return FALSE;

	/* Check Present Bit.. */
	if ( !(Descriptor.Present) )
		return FALSE;

	/* Calc the least privlaged level between CPL and RPL. */
	/* CPL is first 2 bits of CS */
	/* RPL is first 2 bits of dwSegmentSelector */
	dwPrvLevel = max ( (gdwCsSelector & 3), (dwSegmentSelector & 3));

	if (Descriptor.DPL < dwPrvLevel)
		return FALSE;
	
	return TRUE;
}


BOOL
IsAddressValid(DWORD dwSegmentSelector, DWORD dwAddress, int size, int mode)
{
	DWORD dwLimit;
	BOOL fValidAccess=TRUE;
	BYTE Descriptor[8];

	if (mode == MODE_WRITEACCESS) {
		_asm {
			verw	WORD PTR dwSegmentSelector
			jz		Here
			mov		fValidAccess, FALSE
		Here:	
		}
	}else{
		_asm {
			verr	WORD PTR dwSegmentSelector
			jz		There
			mov		fValidAccess, FALSE
	 	There:
		}
	}

	if (!fValidAccess)
		return FALSE;

	_asm {
		lsl		eax, dwSegmentSelector
		mov		dwLimit, eax
	}

	GetDescriptorFields (dwSegmentSelector, &Descriptor);

	/* Expand down segments, the address is valid above the limit, not */
	/* within the limits.. */
	/* */
	if (Descriptor[5] & 0x4) {	
		/* Expand Down Segments:
		 * 	. Just look for start address to be bad..
		 */
		if ( dwAddress <= dwLimit)	/* 0 - limit is bad */
			return FALSE;
		/* if G=0, > 1G-1 is bad*/
		/*if ( ((Descriptor[6] & 0x8) == 0) && (dwAddress > 0xfffff) )	
					return FALSE;
		*/
	} else {
		/* Expand up segments, just look for last address to be bad */
		if ( (dwAddress+size-1) > dwLimit)
			return FALSE;
	}

	return TRUE;
}

BOOL
MapAddressFlat (MMX64 **pPtr, int size, int mode)
{
	DWORD dwSegBaseAddr=0;

    if (gfMmRegUsed)
       return TRUE;

	if ( !IsAddressValid (gdwUsedSegSelector, (DWORD)*pPtr, size, mode) ) {
		ReturnFaultToDebugger(ERR_INVALID_SEGMENT_LIMIT);
		return FALSE;
		}

	if ( !MapSelectorFlat (gdwUsedSegSelector, &dwSegBaseAddr) ) {
		ReturnFaultToDebugger(ERR_SEGMENT_MAPPING_FAILED);
		return FALSE;
		}

	(BYTE *)*pPtr += dwSegBaseAddr;		/* FLAT Linear address.. */

	if (!CheckAddressRange((DWORD)*pPtr, size)) {
		ReturnFaultToDebugger(ERR_PAGENOTPRESENT);
		return FALSE;
		}

	return TRUE;
}
#endif 		/* WIN_VXD */


/* -------- BEGIN WIN_APP ---------------------------------------------*/
#ifdef WIN_APP
int	continue_search;
PEXCEPTION_POINTERS lucp;

void	
fatal(char *msg, int sig)
{
#ifdef  ERROR_BOX
	char lineBuf[256];

	wsprintf(lineBuf,"%s, IP=0x%8.8X, instruction first 2 bytes=0x%2.2X%2.2X",
		msg, (int)R__EIP, *(CHAR *)R__EIP, *(CHAR *)(R__EIP+1));
	MessageBox(NULL, msg, "ERROR", MB_OK);
#endif
}

int MMxHandleException(PEXCEPTION_POINTERS pucp, int exCode)
{
	lucp = pucp;

	/*** copy context upon exception since it's alive in filter only ***/
	/* Fix NT bug of decoding PSARD (0F F0) as a LOCK prefixes (F0)
	0xc000001e is the reserved exception, internally to Microsoft know as
	STATUS_INVALID_LOCK_SEQUENCE. This define can be found in ntstatus.h */
	if ( (exCode == EXCEPTION_ILLEGAL_INSTRUCTION) || (exCode == 0xc000001e) )  {
		CopyMemory(gFpContext, &lucp->ContextRecord->FloatSave, sizeof(FP_CONTEXT));
		if (gFpContext->StatusWord & 0x3800)
			rotate_context();
		fastDecode();
		CopyMemory(&lucp->ContextRecord->FloatSave, gFpContext, sizeof(FP_CONTEXT));
		
		if (gBytesProcessed)
			return(EXCEPTION_CONTINUE_EXECUTION);
	}
		
	return EXCEPTION_CONTINUE_SEARCH;
}
#endif


/* -------- BEGIN UNIX_APP ---------------------------------------------*/
#if !defined(WIN_APP) && !defined(WIN_VXD)
static	ucontext_t	ucp;
static	ucontext_t	*lucp;

static	void	fatal(char *msg, int sig)
{
	fprintf(stderr,"%s, IP=0x%8.8X, instruction first 2 bytes=0x%2.2X%2.2X\n",
		msg, (int)R__EIP, *(uchar *)R__EIP, *(uchar *)(R__EIP+1));
	exit(sig);
}

extern  DWORD gSize;
static	void	emulate_mmx(int i)
{
	(void)getcontext(&ucp);
	lucp=ucp.uc_link;
	
	memcpy(gFpContext, &lucp->uc_mcontext.fpregs.fp_reg_set.fpchip_state.state[0], sizeof(FP_CONTEXT));
	if (gFpContext->StatusWord & 0x3800)
		rotate_context();

	decode();
	memcpy(&lucp->uc_mcontext.fpregs.fp_reg_set.fpchip_state.state[0], gFpContext, sizeof(FP_CONTEXT));

	lucp->uc_mcontext.gregs[EIP] += gSize;
	setcontext(lucp);
}

/* This is for backward compatibility with older releases of the simualtor.. you should
 * use start_mmx() from now on..
 */
start_simd()
{
  	signal(SIGILL, emulate_mmx);
}

start_mmx()
{
  	signal(SIGILL, emulate_mmx);
}

#endif	/* UNIX APP */




#ifdef WIN_VXD
typedef struct _MMX_INFO
{
	DWORD   dwCsReg;
	DWORD	dwEipReg;
	DWORD 	mmReg[8][2];			/* 8 regs, 2DWORDS each */
	WORD	nOpcode;
	BYTE	Opcode[20];
}MMX_INFO;

void ReadMmxInfo (MMX_INFO *pInfo)
{
	int i;

	/* Note that we don't use the FP registers to get the values
	 * we just display what's in the pMmReg memory location.
	 * From the VxD, we have no way of getting the context of the
	 * process or thread that is running at the time.
	 */
	
	pInfo->nOpcode = gLastOpcodeSize;

	for (i=0; i<gLastOpcodeSize; i++)
		pInfo->Opcode[i] = gLastOpcode[i];

	pInfo->dwCsReg = gdwLastCsReg;
	pInfo->dwEipReg = gdwLastEipReg;

	for (i=0; i<8; i++) {
		pInfo->mmReg[i][0] = pMmReg[i].Mantissa.v.d[0];
		pInfo->mmReg[i][1] = pMmReg[i].Mantissa.v.d[1];
	}
}

void WriteMmxInfo (DWORD *pReg, DWORD nReg)
{
	/* Note that we don't use the FP registers to get the values
	 * we just display what's in the pMmReg memory location.
	 * From the VxD, we have no way of getting the context of the
	 * process or thread that is running at the time.
	 */

	pMmReg[nReg].Mantissa.v.d[0] = pReg[0];
	pMmReg[nReg].Mantissa.v.d[1] = pReg[1];
}

char gszTmp[1024];
DisplayErrorBox (char *szMsg)
{
	int i=0;
	int j=0;

	for (i=0; szMsg[i]; i++, j++)
		gszTmp[j] = szMsg[i];

	for (i=0; i<sizeof(ERR_GENERAL); i++, j++)
		gszTmp[j] = ERR_GENERAL[i];

	gszTmp[j] = '\0';

	DisplayMessage(gszTmp);
}

#endif


/************** Data Move Instructions *************************************/
/*-------------------------------------------------------------------------*/
static void movd_st ( MMX64 *dst, MMX64 *src , int x86reg ,int dst_index)
{	unsigned long val;

	val =  src->v.d[0];
	if (x86reg)
		switch (dst_index) 
		{
			case 0: R__EAX  = val; break;
			case 1: R__ECX  = val; break;
			case 2: R__EDX  = val; break;
			case 3: R__EBX  = val; break;
			case 4: R__ESP  = val; break;
			case 5: R__EBP  = val; break;
			case 6: R__ESI  = val; break;
			case 7: R__EDI  = val; break;
		}
	else {
		if (!MapAddressFlat(&dst, 4, MODE_WRITEACCESS) )
			return;
		dst->v.d[0] = val;
		}
}

/*-------------------------------------------------------------------------*/
static void movd_ld ( MMX64 *src, MMX64 *dst , int x86reg ,int src_index)
{	unsigned long val;
	if (x86reg)
		switch (src_index) 
		{
			case 0:  val = R__EAX; break;
			case 1:  val = R__ECX; break;
			case 2:  val = R__EDX; break;
			case 3:  val = R__EBX; break;
			case 4:  val = R__ESP; break;
			case 5:  val = R__EBP; break;
			case 6:  val = R__ESI; break;
			case 7:  val = R__EDI; break;
		}
	else {
		if (!MapAddressFlat(&src, 4, MODE_READACCESS) )
			return;
		val =  src->v.d[0];
		}
	dst->v.d[0] = val;
	dst->v.d[1] = 0;
}

/*-------------------------------------------------------------------------*/
static void movq ( MMX64 *src, MMX64 *dst)
{ 
	*dst = *src;
}

#ifdef WIN_VXD
static void rdpmc()
{
	DWORD dwEcx = R__ECX;
	DWORD dwEdx, dwEax;

	/* execute RDMSR previlaged instruction, to get counter0,1. */
	/* These counters are index 0x12,0x13 for RDMSR. */
	/* RDPMC expects the results in EDX:EAX */
	/* */
	if (dwEcx == 0)
		_asm mov ecx, 0x12;
	else if (dwEcx == 1)
		_asm mov ecx, 0x13;
	else {
		DisplayErrorBox(ERR_RDPMC_INVALID_PARM); 
		return;
	}

	_asm {
		_emit 0x0f		; RDMSR opcode
	 	_emit 0x32
		mov	dwEdx, edx
		mov dwEax, eax
	}
	
	R__EDX = dwEdx;
	R__EAX = dwEax;
}
#endif 

#include "mmx.h"

/************** General purpose macros ********************************/
/*
 * returns EX86's cmp-val: 
 * 0x0 sim_or 0xffffffff for C's FALSE sim_or TRUE (respectively)
 */
#define CV(xx)  ((long)((xx) ? (-1) : (0)))

/************** Saturation macros *************************************/

/* saturate_? signed/unsigned saturate a number acoording to it's size */
#define SATURATE8_S(res)                                   \
{                                                          \
        if ( (res) >= 128 )                                \
            (res) = 127;                                   \
        else if ( (res) < -128 )                           \
                (res) = -128;                              \
}

#define SATURATE16_S(res)                                  \
{                                                          \
        if ( (res) >= 32768 )                              \
            (res) = 32767;                                 \
        else if ( (res) < -32768 )                         \
                (res) = -32768;                            \
}

#define SATURATE8_U(res)                                   \
{                                                          \
        if ( (res) >= 256 )                                \
            (res) = 255;                                   \
        else if ( (res) < 0 )                              \
                (res) = 0;                                 \
}

#define SATURATE16_U(res)                                  \
{                                                          \
        if ( (res) >= 65536 )                              \
            (res) = 65535;                                 \
        else if ( (res) < 0 )                              \
                (res) = 0;                                 \
}

/********************************************************************/


/*---------------------------------------------------------------------------
 *  packsswb -- 
 *---------------------------------------------------------------------------
 */
void packsswb(MMX64 *dst, MMX64 *src)
{
	int   i;
	long  t; 
	MMX64 tmp;

	for (i=0; i<4; i++)
	{
		t = (long)dst->v.ws[i];
		SATURATE8_S(t);
		tmp.v.bs[i] = (MMXS8)t;
		t = (long)src->v.ws[i];
		SATURATE8_S(t);
		tmp.v.bs[i+4] = (MMXS8)t;
	}
	*dst = tmp;
}


/*---------------------------------------------------------------------------
 *  packssdw -- 
 *---------------------------------------------------------------------------
 */
void packssdw(MMX64 *dst, MMX64 *src)
{
	int   i;
	long  t; 
	MMX64 tmp;

	for (i=0; i<2; i++)
	{
		t = (long)dst->v.ds[i];
		SATURATE16_S(t);
		tmp.v.ws[i] = (MMXS16)t;
		t = (long)src->v.ds[i];
		SATURATE16_S(t);
		tmp.v.ws[i+2] = (MMXS16)t;
	}
	*dst = tmp;
}


/*---------------------------------------------------------------------------
 *  packuswb -- 
 *---------------------------------------------------------------------------
 */
void packuswb(MMX64 *dst, MMX64 *src)
{
	int   i;
	long  t; 
	MMX64 tmp;

	for (i=0; i<4; i++)
	{
		t = (long)dst->v.ws[i];
		SATURATE8_U(t);
		tmp.v.b[i] = (MMXU8)t;
		t = (long)src->v.ws[i];
		SATURATE8_U(t);
		tmp.v.b[i+4] = (MMXU8)t;
	}
	*dst = tmp;
}


/*---------------------------------------------------------------------------
 *  paddb -- 
 *---------------------------------------------------------------------------
 */
void paddb(MMX64 *dst, MMX64 *src)
{
	int i;

	for (i=0; i<8; i++)
	{
		dst->v.b[i] += src->v.b[i];
	}
}


/*---------------------------------------------------------------------------
 *  paddw -- 
 *---------------------------------------------------------------------------
 */
void paddw(MMX64 *dst, MMX64 *src)
{
	int i;

	for (i=0; i<4; i++)
	{
		dst->v.w[i] += src->v.w[i];
	}
}


/*---------------------------------------------------------------------------
 *  paddd -- 
 *---------------------------------------------------------------------------
 */
void paddd(MMX64 *dst, MMX64 *src)
{
	int i;

	for (i=0; i<2; i++)
	{
		dst->v.d[i] += src->v.d[i];
	}
}


/*---------------------------------------------------------------------------
 *  paddsb -- 
 *---------------------------------------------------------------------------
 */
void paddsb(MMX64 *dst, MMX64 *src)
{
	int  i;
	long t; 

	for (i=0; i<8; i++)
	{
		t = (long)(dst->v.bs[i] + src->v.bs[i]);
		SATURATE8_S(t);
		dst->v.bs[i] = (MMXS8)t;
	}
}


/*---------------------------------------------------------------------------
 *  paddsw -- 
 *---------------------------------------------------------------------------
 */
void paddsw(MMX64 *dst, MMX64 *src)
{
	int  i;
	long t; 

	for (i=0; i<4; i++)
	{
		t = (long)(dst->v.ws[i] + src->v.ws[i]);
		SATURATE16_S(t);
		dst->v.ws[i] = (MMXS16)t;
	}
}


/*---------------------------------------------------------------------------
 *  paddusb -- 
 *---------------------------------------------------------------------------
 */
void paddusb(MMX64 *dst, MMX64 *src)
{
	int  i;
	long t; 

	for (i=0; i<8; i++)
	{
		t = (long)(dst->v.b[i] + src->v.b[i]);
		SATURATE8_U(t);
		dst->v.b[i] = (MMXU8)t;
	}
}


/*---------------------------------------------------------------------------
 *  paddusw -- 
 *---------------------------------------------------------------------------
 */
void paddusw(MMX64 *dst, MMX64 *src)
{
	int  i;
	long t; 

	for (i=0; i<4; i++)
	{
		t = (long)(dst->v.w[i] + src->v.w[i]);
		SATURATE16_U(t);
		dst->v.w[i] = (MMXU16)t;
	}
}


/*---------------------------------------------------------------------------
 *  pand -- 
 *---------------------------------------------------------------------------
 */
void pand(MMX64 *dst, MMX64 *src)
{
	dst->v.d[0] &= src->v.d[0];
	dst->v.d[1] &= src->v.d[1];
}


/*---------------------------------------------------------------------------
 *  pandn -- 
 *---------------------------------------------------------------------------
 */
void pandn(MMX64 *dst, MMX64 *src)
{
	dst->v.d[0] = (~dst->v.d[0]) & src->v.d[0];
	dst->v.d[1] = (~dst->v.d[1]) & src->v.d[1];
}


/*---------------------------------------------------------------------------
 *  pcmpeqb -- 
 *---------------------------------------------------------------------------
 */
void pcmpeqb(MMX64 *dst, MMX64 *src)
{
	int  i;

	for (i=0; i<8; i++)
	{
		dst->v.bs[i] = (MMXS8)CV(src->v.bs[i] == dst->v.bs[i]);
	}
}


/*---------------------------------------------------------------------------
 *  pcmpeqw -- 
 *---------------------------------------------------------------------------
 */
void pcmpeqw(MMX64 *dst, MMX64 *src)
{
	int  i;

	for (i=0; i<4; i++)
	{
		dst->v.ws[i] = (MMXS16)CV(src->v.ws[i] == dst->v.ws[i]);
	}
}


/*---------------------------------------------------------------------------
 *  pcmpeqd -- 
 *---------------------------------------------------------------------------
 */
void pcmpeqd(MMX64 *dst, MMX64 *src)
{
	int  i;

	for (i=0; i<2; i++)
	{
		dst->v.ds[i] = (MMXS32)CV(src->v.ds[i] == dst->v.ds[i]);
	}
}


/*---------------------------------------------------------------------------
 *  pcmpgtb -- 
 *---------------------------------------------------------------------------
 */
void pcmpgtb(MMX64 *dst, MMX64 *src)
{
	int  i;

	for (i=0; i<8; i++)
	{
		dst->v.bs[i] = (MMXS8)CV(src->v.bs[i] < dst->v.bs[i]);
	}
}


/*---------------------------------------------------------------------------
 *  pcmpgtw -- 
 *---------------------------------------------------------------------------
 */
void pcmpgtw(MMX64 *dst, MMX64 *src)
{
	int  i;

	for (i=0; i<4; i++)
	{
		dst->v.ws[i] = (MMXS16)CV(src->v.ws[i] < dst->v.ws[i]);
	}
}


/*---------------------------------------------------------------------------
 *  pcmpgtd -- 
 *---------------------------------------------------------------------------
 */
void pcmpgtd(MMX64 *dst, MMX64 *src)
{
	int  i;

	for (i=0; i<2; i++)
	{
		dst->v.ds[i] = (MMXS32)CV(src->v.ds[i] < dst->v.ds[i]);
	}
}


/*---------------------------------------------------------------------------
 *  pmaddwd -- 
 *---------------------------------------------------------------------------
 */
void pmaddwd(MMX64 *dst, MMX64 *src)
{
	dst->v.ds[0] = (dst->v.ws[0] * src->v.ws[0]) +
	              (dst->v.ws[1] * src->v.ws[1]);
	dst->v.ds[1] = (dst->v.ws[2] * src->v.ws[2]) +
	              (dst->v.ws[3] * src->v.ws[3]);
}


/*---------------------------------------------------------------------------
 *  pmulhw -- 
 *---------------------------------------------------------------------------
 */
void pmulhw(MMX64 *dst, MMX64 *src)
{
	int  i;
	long t;

	for (i=0; i<4; i++)
	{
		t = dst->v.ws[i] * src->v.ws[i];
		t >>= 16;    
		dst->v.ws[i] = (MMXS16)t;
	}
}


/*---------------------------------------------------------------------------
 *  pmullw -- 
 *---------------------------------------------------------------------------
 */
void pmullw(MMX64 *dst, MMX64 *src)
{
	int  i;

	for (i=0; i<4; i++)
	{
		dst->v.ws[i] *= src->v.ws[i];
	}
}


/*---------------------------------------------------------------------------
 *  por -- 
 *---------------------------------------------------------------------------
 */
void por(MMX64 *dst, MMX64 *src)
{
	dst->v.d[0] |= src->v.d[0];
	dst->v.d[1] |= src->v.d[1];
}


/*---------------------------------------------------------------------------
 *  psllw -- 
 *---------------------------------------------------------------------------
 */
void psllw(MMX64 *dst, MMX64 *src)
{
	if ( src->v.d[1] || src->v.d[0] > 15)
	{
		dst->v.w[0] = 0;
		dst->v.w[1] = 0;
		dst->v.w[2] = 0;
		dst->v.w[3] = 0;
	}
	else
	{	
		dst->v.w[0] <<= src->v.d[0];
		dst->v.w[1] <<= src->v.d[0];
		dst->v.w[2] <<= src->v.d[0];
		dst->v.w[3] <<= src->v.d[0];
	}
}


/*---------------------------------------------------------------------------
 *  pslld -- 
 *---------------------------------------------------------------------------
 */
void pslld(MMX64 *dst, MMX64 *src)
{
	if ( src->v.d[1] || src->v.d[0] > 31)
	{
		dst->v.d[0] = 0;
		dst->v.d[1] = 0;
	}
	else
	{	
		dst->v.d[0] <<= src->v.d[0];
		dst->v.d[1] <<= src->v.d[0];
	}
}


/*---------------------------------------------------------------------------
 *  psllq -- 
 *---------------------------------------------------------------------------
 */
void psllq(MMX64 *dst, MMX64 *src)
{
	long bm;

	if ( src->v.d[1] || src->v.d[0] > 63)
	{
		dst->v.d[0] = 0;
		dst->v.d[1] = 0;
	}
	else if ( src->v.d[0] == 0 )
	{
		/* EMPTY - shift by zero, destination does not change */
	}
	else
	{
		bm = src->v.d[0] - 32;
		if (bm >= 0)
		{
			dst->v.d[1] = dst->v.d[0] << bm;
			dst->v.d[0]  = 0;
		}
		else
		{
			dst->v.d[1]  = (dst->v.d[1] << src->v.d[0]) |
			  (dst->v.d[0]  >> (-bm));
			dst->v.d[0]  =  dst->v.d[0] << src->v.d[0];
		}	
	}
}


/*---------------------------------------------------------------------------
 *  psllw_imm8 -- 
 *---------------------------------------------------------------------------
 */
void psllw_imm8(MMX64 *dst, MMXU8 imm8)
{
	if ( imm8 > 15)
	{
		dst->v.w[0] = 0;
		dst->v.w[1] = 0;
		dst->v.w[2] = 0;
		dst->v.w[3] = 0;
	}
	else
	{	
		dst->v.w[0] <<= imm8;
		dst->v.w[1] <<= imm8;
		dst->v.w[2] <<= imm8;
		dst->v.w[3] <<= imm8;
	}
}


/*---------------------------------------------------------------------------
 *  pslld_imm8 -- 
 *---------------------------------------------------------------------------
 */
void pslld_imm8(MMX64 *dst, MMXU8 imm8)
{
	if ( imm8 > 31)
	{
		dst->v.d[0] = 0;
		dst->v.d[1] = 0;
	}
	else
	{	
		dst->v.d[0] <<= imm8;
		dst->v.d[1] <<= imm8;
	}
}


/*---------------------------------------------------------------------------
 *  psllq_imm8 -- 
 *---------------------------------------------------------------------------
 */
void psllq_imm8(MMX64 *dst, MMXU8 imm8)
{
	long bm;

	if ( imm8 > 63)
	{
		dst->v.d[0] = 0;
		dst->v.d[1] = 0;
	}
	else if ( imm8 == 0 )
	{
		/* EMPTY - shift by zero, destination does not change */
	}
	else
	{
		bm = imm8 - 32;
		if (bm >= 0)
		{
			dst->v.d[1] = dst->v.d[0] << bm;
			dst->v.d[0]  = 0;
		}
		else
		{
			dst->v.d[1]  = (dst->v.d[1] << imm8) |
			  (dst->v.d[0]  >> (32-imm8));
			dst->v.d[0]  =  dst->v.d[0] << imm8;
		}
	}
}


/*---------------------------------------------------------------------------
 *  psraw -- 
 *---------------------------------------------------------------------------
 */
void psraw(MMX64 *dst, MMX64 *src)
{
	if ( src->v.d[1] || src->v.d[0] > 15)
	{
		dst->v.ws[0] = (dst->v.ws[0] & 0x8000) ? -1 : 0;
		dst->v.ws[1] = (dst->v.ws[1] & 0x8000) ? -1 : 0;
		dst->v.ws[2] = (dst->v.ws[2] & 0x8000) ? -1 : 0;
		dst->v.ws[3] = (dst->v.ws[3] & 0x8000) ? -1 : 0;
	}
	else
	{	
		dst->v.ws[0] >>= src->v.ds[0];
		dst->v.ws[1] >>= src->v.ds[0];
		dst->v.ws[2] >>= src->v.ds[0];
		dst->v.ws[3] >>= src->v.ds[0];
	}
}


/*---------------------------------------------------------------------------
 *  psrad -- 
 *---------------------------------------------------------------------------
 */
void psrad(MMX64 *dst, MMX64 *src)
{
	if ( src->v.d[1] || src->v.d[0] > 31)
	{
		dst->v.ds[0] = (dst->v.ds[0] & 0x80000000) ? -1 : 0;
		dst->v.ds[1] = (dst->v.ds[1] & 0x80000000) ? -1 : 0;
	}
	else
	{	
		dst->v.ds[0] >>= src->v.ds[0];
		dst->v.ds[1] >>= src->v.ds[0];
	}
}


/*---------------------------------------------------------------------------
 *  psraw_imm8 -- 
 *---------------------------------------------------------------------------
 */
void psraw_imm8(MMX64 *dst, MMXU8 imm8)
{
	if ( imm8 > 15)
	{
		dst->v.ws[0] = (dst->v.ws[0] & 0x8000) ? -1 : 0;
		dst->v.ws[1] = (dst->v.ws[1] & 0x8000) ? -1 : 0;
		dst->v.ws[2] = (dst->v.ws[2] & 0x8000) ? -1 : 0;
		dst->v.ws[3] = (dst->v.ws[3] & 0x8000) ? -1 : 0;
	}
	else
	{	
		dst->v.ws[0] >>= (MMXS8)imm8;
		dst->v.ws[1] >>= (MMXS8)imm8;
		dst->v.ws[2] >>= (MMXS8)imm8;
		dst->v.ws[3] >>= (MMXS8)imm8;
	}
}


/*---------------------------------------------------------------------------
 *  psrad_imm8 -- 
 *---------------------------------------------------------------------------
 */
void psrad_imm8(MMX64 *dst, MMXU8 imm8)
{
	if ( imm8 > 31)
	{
		dst->v.ds[0] = (dst->v.ds[0] & 0x80000000) ? -1 : 0;
		dst->v.ds[1] = (dst->v.ds[1] & 0x80000000) ? -1 : 0;
	}
	else
	{	
		dst->v.ds[0] >>= (MMXS8)imm8;
		dst->v.ds[1] >>= (MMXS8)imm8;
	}
}


/*---------------------------------------------------------------------------
 *  psrlw -- 
 *---------------------------------------------------------------------------
 */
void psrlw(MMX64 *dst, MMX64 *src)
{
	if ( src->v.d[1] || src->v.d[0] > 15)
	{
		dst->v.w[0] = 0;
		dst->v.w[1] = 0;
		dst->v.w[2] = 0;
		dst->v.w[3] = 0;
	}
	else
	{	
		dst->v.w[0] >>= src->v.d[0];
		dst->v.w[1] >>= src->v.d[0];
		dst->v.w[2] >>= src->v.d[0];
		dst->v.w[3] >>= src->v.d[0];
	}
}


/*---------------------------------------------------------------------------
 *  psrld -- 
 *---------------------------------------------------------------------------
 */
void psrld(MMX64 *dst, MMX64 *src)
{
	if ( src->v.d[1] || src->v.d[0] > 31)
	{
		dst->v.d[0] = 0;
		dst->v.d[1] = 0;
	}
	else
	{	
		dst->v.d[0] >>= src->v.d[0];
		dst->v.d[1] >>= src->v.d[0];
	}
}


/*---------------------------------------------------------------------------
 *  psrlq -- 
 *---------------------------------------------------------------------------
 */
void psrlq(MMX64 *dst, MMX64 *src)
{
	long bm;

	if ( src->v.d[1] || src->v.d[0] > 63)
	{
		dst->v.d[0] = 0;
		dst->v.d[1] = 0;
	}
	else if ( src->v.d[0] == 0 )
	{
		/* EMPTY - shift by zero, destination does not change */
	}
	else
	{
		bm = src->v.d[0] - 32;
		if (bm >= 0)
		{
			dst->v.d[0] = dst->v.d[1] >> bm;
			dst->v.d[1]  = 0;
		}
		else
		{
			dst->v.d[0]  = (dst->v.d[0] >> src->v.d[0]) |
			  (dst->v.d[1]  << (-bm));
			dst->v.d[1]  =  dst->v.d[1] >> src->v.d[0];
		}	
	}
}


/*---------------------------------------------------------------------------
 *  psrlw_imm8 -- 
 *---------------------------------------------------------------------------
 */
void psrlw_imm8(MMX64 *dst, MMXU8 imm8)
{
	if ( imm8 > 15)
	{
		dst->v.w[0] = 0;
		dst->v.w[1] = 0;
		dst->v.w[2] = 0;
		dst->v.w[3] = 0;
	}
	else
	{	
		dst->v.w[0] >>= imm8;
		dst->v.w[1] >>= imm8;
		dst->v.w[2] >>= imm8;
		dst->v.w[3] >>= imm8;
	}
}


/*---------------------------------------------------------------------------
 *  psrld_imm8 -- 
 *---------------------------------------------------------------------------
 */
void psrld_imm8(MMX64 *dst, MMXU8 imm8)
{
	if ( imm8 > 31)
	{
		dst->v.d[0] = 0;
		dst->v.d[1] = 0;
	}
	else
	{	
		dst->v.d[0] >>= imm8;
		dst->v.d[1] >>= imm8;
	}
}


/*---------------------------------------------------------------------------
 *  psrlq_imm8 -- 
 *---------------------------------------------------------------------------
 */
void psrlq_imm8(MMX64 *dst, MMXU8 imm8)
{
	long bm;

	if ( imm8 > 63)
	{
		dst->v.d[0] = 0;
		dst->v.d[1] = 0;
	}
	else if ( imm8 == 0 )
	{
		/* EMPTY - shift by zero, destination does not change */
	}
	else
	{
		bm = imm8 - 32;
		if (bm >= 0)
		{
			dst->v.d[0] = dst->v.d[1] >> bm;
			dst->v.d[1]  = 0;
		}
		else
		{
			dst->v.d[0]  = (dst->v.d[0] >> imm8) |
			  (dst->v.d[1]  << (32-imm8));
			dst->v.d[1]  =  dst->v.d[1] >> imm8;
		}
	}
}


/*---------------------------------------------------------------------------
 *  psubb -- 
 *---------------------------------------------------------------------------
 */
void psubb(MMX64 *dst, MMX64 *src)
{
	int i;

	for (i=0; i<8; i++)
	{
		dst->v.b[i] -= src->v.b[i];
	}
}


/*---------------------------------------------------------------------------
 *  psubw -- 
 *---------------------------------------------------------------------------
 */
void psubw(MMX64 *dst, MMX64 *src)
{
	int i;

	for (i=0; i<4; i++)
	{
		dst->v.w[i] -= src->v.w[i];
	}
}


/*---------------------------------------------------------------------------
 *  psubd -- 
 *---------------------------------------------------------------------------
 */
void psubd(MMX64 *dst, MMX64 *src)
{
	int i;

	for (i=0; i<2; i++)
	{
		dst->v.d[i] -= src->v.d[i];
	}
}


/*---------------------------------------------------------------------------
 *  psubsb -- 
 *---------------------------------------------------------------------------
 */
void psubsb(MMX64 *dst, MMX64 *src)
{
	int  i;
	long t; 

	for (i=0; i<8; i++)
	{
		t = (long)(dst->v.bs[i] - src->v.bs[i]);
		SATURATE8_S(t);
		dst->v.bs[i] = (MMXS8)t;
	}
}


/*---------------------------------------------------------------------------
 *  psubsw -- 
 *---------------------------------------------------------------------------
 */
void psubsw(MMX64 *dst, MMX64 *src)
{
	int  i;
	long t; 

	for (i=0; i<4; i++)
	{
		t = (long)(dst->v.ws[i] - src->v.ws[i]);
		SATURATE16_S(t);
		dst->v.ws[i] = (MMXS16)t;
	}
}


/*---------------------------------------------------------------------------
 *  psubusb -- 
 *---------------------------------------------------------------------------
 */
void psubusb(MMX64 *dst, MMX64 *src)
{
	int  i;
	long t; 

	for (i=0; i<8; i++)
	{
		t = (long)(dst->v.b[i] - src->v.b[i]);
		SATURATE8_U(t);
		dst->v.b[i] = (MMXU8)t;
	}
}


/*---------------------------------------------------------------------------
 *  psubusw -- 
 *---------------------------------------------------------------------------
 */
void psubusw(MMX64 *dst, MMX64 *src)
{
	int  i;
	long t; 

	for (i=0; i<4; i++)
	{
		t = (long)(dst->v.w[i] - src->v.w[i]);
		SATURATE16_U(t);
		dst->v.w[i] = (MMXU16)t;
	}
}


/*---------------------------------------------------------------------------
 *  punpckhbw -- 
 *---------------------------------------------------------------------------
 */
void punpckhbw(MMX64 *dst, MMX64 *src)
{
    MMX64 tmp;

	tmp.v.b[0] = dst->v.b[4];
	tmp.v.b[1] = src->v.b[4];
	tmp.v.b[2] = dst->v.b[5];
	tmp.v.b[3] = src->v.b[5];
	tmp.v.b[4] = dst->v.b[6];
	tmp.v.b[5] = src->v.b[6];
	tmp.v.b[6] = dst->v.b[7];
	tmp.v.b[7] = src->v.b[7];
	*dst = tmp;
}


/*---------------------------------------------------------------------------
 *  punpckhwd -- 
 *---------------------------------------------------------------------------
 */
void punpckhwd(MMX64 *dst, MMX64 *src)
{
    MMX64 tmp;

	tmp.v.w[0] = dst->v.w[2];
	tmp.v.w[1] = src->v.w[2];
	tmp.v.w[2] = dst->v.w[3];
	tmp.v.w[3] = src->v.w[3];
	*dst = tmp;
}


/*---------------------------------------------------------------------------
 *  punpckhdq -- 
 *---------------------------------------------------------------------------
 */
void punpckhdq(MMX64 *dst, MMX64 *src)
{
    MMX64 tmp;

	tmp.v.d[0] = dst->v.d[1];
	tmp.v.d[1] = src->v.d[1];
	*dst = tmp;
}


/*---------------------------------------------------------------------------
 *  punpcklbw -- 
 *---------------------------------------------------------------------------
 */
void punpcklbw(MMX64 *dst, MMX64 *src)
{
    MMX64 tmp;

	tmp.v.b[0] = dst->v.b[0];
	tmp.v.b[1] = src->v.b[0];
	tmp.v.b[2] = dst->v.b[1];
	tmp.v.b[3] = src->v.b[1];
	tmp.v.b[4] = dst->v.b[2];
	tmp.v.b[5] = src->v.b[2];
	tmp.v.b[6] = dst->v.b[3];
	tmp.v.b[7] = src->v.b[3];
	*dst = tmp;
}


/*---------------------------------------------------------------------------
 *  punpcklwd -- 
 *---------------------------------------------------------------------------
 */
void punpcklwd(MMX64 *dst, MMX64 *src)
{
    MMX64 tmp;

	tmp.v.w[0] = dst->v.w[0];
	tmp.v.w[1] = src->v.w[0];
	tmp.v.w[2] = dst->v.w[1];
	tmp.v.w[3] = src->v.w[1];
	*dst = tmp;
}


/*---------------------------------------------------------------------------
 *  punpckldq -- 
 *---------------------------------------------------------------------------
 */
void punpckldq(MMX64 *dst, MMX64 *src)
{
    MMX64 tmp;

	tmp.v.d[0] = dst->v.d[0];
	tmp.v.d[1] = src->v.d[0];
	*dst = tmp;
}


/*---------------------------------------------------------------------------
 *  pxor -- 
 *---------------------------------------------------------------------------
 */
void pxor(MMX64 *dst, MMX64 *src)
{
	dst->v.d[0] ^= src->v.d[0];
	dst->v.d[1] ^= src->v.d[1];
}
