/*
*    INTEL CORPORATION PROPRIETARY INFORMATION 
*
*    This software is supplied under the terms of a license
*    agreement or nondisclosure agreement with Intel Corporation 
*    and may not be copied or disclosed except in accordance with 
*    the terms of that agreement.
*    Copyright (c) 1991,1992,1993  Intel Corporation. 
*
*/

typedef unsigned char uchar;
typedef signed char signed_char;
typedef unsigned long   U32;       /* type of 32 bit items */

#ifndef TRUE
   #define TRUE 1
   #define FALSE 0
#endif
#ifndef BOOL
  typedef int BOOL;
#endif

#ifndef WORD
   typedef unsigned short WORD;
#endif

#ifndef DWORD
   typedef unsigned long DWORD;
#endif

enum { MODE_WRITEACCESS, MODE_READACCESS };


typedef struct { union {
					U32 d[2]; 
					uchar b[8];
					unsigned short w[4];
					} v;
				} U64;  /* type of 64 bit items */

#define LOW_32 0
#define HI_32 1
/*
* returns  cmp-val: 
* 0x0 sim_or 0xffffffff for C's FALSE sim_or TRUE (respectively)
*/

/************** Macros to access register file ************************/


/* Fixed point registers */
#define CLEAN(reg)  reg.v.d[0] = reg.v.d[1] = 0

#define bXR(ri,wi)   ri->v.b[wi]
#define wXR(ri,wi)   ri->v.w[wi]
#define R(ri,wi) (ri->v.d[wi])
#define SR(ri,wi) ((signed long) (ri->v.d[wi]))
#define UR(ri,wi) ((unsigned long) (ri->v.d[wi]))
/* assign a value to a specific byte/word (U64 t)*/

#define bXSET(t,wi,val)   t.v.b[wi]=(uchar)val
#define wXSET(t,wi,val)   t.v.w[wi]=(unsigned short)val
#define dXSET(t,wi,val)   t.v.d[wi]=(U32)val

/************** Saturation macros *************************************/

/* saturate_? signed/unsigned saturate a number acoording to it's size */
#define SATURATE_S(res,sz)                                 \
{                                                          \
        if ( (res) >= ((long)1 << ((sz)-1)) )              \
            (res) = ((long)1 <<((sz)-1))-1;                \
        else if ((res)< -((long)1 << ((sz)-1)))            \
                (res) = -((long)1 << ((sz)-1));            \
}

#define SATURATE_U(res,sz)                                  \
{                                                           \
        if ( (res) >= ((long)1 << (sz)) )                   \
            (res) = ((long)1 <<(sz))-1;                     \
        else if ((res)<0)                                   \
                (res) = 0;                                  \
}



extern char lineBuf[];
BOOL decode();

#pragma pack(1)
typedef struct { 
	U64 Mantissa;  			   		/* 64 bit Mantissa */
	WORD Exponent;					/* 16 bit Sign and Exponent */
} U80;

typedef struct tagFP_CONTEXT {
	DWORD 	ControlWord;
	DWORD	StatusWord;
	DWORD	TagWord;
	DWORD	IPOffset;
	WORD	CsSelector;
	WORD	OpCode;
	DWORD	OperandOffset;
	DWORD	OperandSelector;
	U80		FpRegisters[8];			/* 8 FP regs, 80 bits each. */
}FP_CONTEXT;

#pragma pack()

/* ++++++++ BEGIN WIN_VXD ++++++++++++++++++++++++++++++++++++++++++++++*/
/* ++++++++ BEGIN WIN_VXD ++++++++++++++++++++++++++++++++++++++++++++++*/
#ifdef WIN_VXD
	#include <windef.h>


//******************************************************************************
// The following structures are pointed to by EBP when VxD routines are entered,
//  both for VxD control calls and traps(I/O traps, software INT traps, etc.).
//  The first structure as DWORD values, the second WORD values and the last
//  has BYTE values.
//******************************************************************************
#pragma pack(1)
typedef struct tagClient_Reg_Struc {
    DWORD	Client_EDI;			/* Client's EDI */
    DWORD	Client_ESI;			/* Client's ESI */
    DWORD	Client_EBP;			/* Client's EBP */
    DWORD	Client_res0;		/* ESP at pushall */
    DWORD	Client_EBX;			/* Client's EBX */
    DWORD	Client_EDX;			/* Client's EDX */
    DWORD	Client_ECX;			/* Client's ECX */
    DWORD	Client_EAX;			/* Client's EAX */
    DWORD	Client_Error; 		/* Dword error code */
    DWORD	Client_EIP;			/* EIP */
    WORD	Client_CS;			/* CS */
    WORD	Client_res1; 		/*   (padding) */
    DWORD	Client_EFlags;		/* EFLAGS */
    DWORD	Client_ESP;			/* ESP */
    WORD	Client_SS;			/* SS */
    WORD	Client_res2; 		/*   (padding) */
    WORD	Client_ES;			/* ES */
    WORD	Client_res3; 		/*   (padding) */
    WORD	Client_DS;			/* DS */
    WORD	Client_res4; 		/*   (padding) */
    WORD	Client_FS;			/* FS */
    WORD	Client_res5; 		/*   (padding) */
    WORD	Client_GS;			/* GS */
    WORD	Client_res6; 		/*   (padding) */
    DWORD	Client_Alt_EIP;
    WORD	Client_Alt_CS;
    WORD	Client_res7;
    DWORD	Client_Alt_EFlags;
    DWORD	Client_Alt_ESP;
    WORD	Client_Alt_SS;
    WORD	Client_res8;
    WORD	Client_Alt_ES;
    WORD	Client_res9;
    WORD	Client_Alt_DS;
    WORD	Client_res10;
    WORD	Client_Alt_FS;
    WORD	Client_res11;
    WORD	Client_Alt_GS;
    WORD	Client_res12;
} Client_Reg_Struc;



typedef struct tagSegDescriptor_t
{
	WORD 	SegLimitLow;
	WORD	BaseAddrLow;
	BYTE	BaseAddrMid;
	BYTE	Type:4;
	BYTE	SBit:1;
	BYTE	DPL:2;
	BYTE	Present:1;
	BYTE	SegLimitHigh:4;
	BYTE	MiscBits:4;
	BYTE	BaseAddrHigh;
}SegDescriptor_t;
#pragma pack()

	/* NOTE: Must have these for a VXD..  VxD's expect to have their code
	 * and data in specific segments, so make sure that all C code does
	 * this.. (well, if you want it to work). 
	 */
//	#pragma code_seg ("_LTEXT")
//	#pragma data_seg ("_LDATA")

	/* Defines ---------------------------------- */
	#define R__EAX		pClientRegs->Client_EAX
	#define R__ECX		pClientRegs->Client_ECX
	#define R__EDX		pClientRegs->Client_EDX
	#define R__EBX		pClientRegs->Client_EBX
	#define R__ESP		pClientRegs->Client_ESP
	#define R__EBP		pClientRegs->Client_EBP
	#define R__ESI		pClientRegs->Client_ESI
	#define R__EDI		pClientRegs->Client_EDI
	#define R__EIP		gpCurEip

	/* Globals ----------------------------------*/
	Client_Reg_Struc *pClientRegs;  /* this is set when this proc is called.*/
	DWORD 	gpCurEip;
	DWORD 	gdwFlatDsSelectorBase;
	DWORD 	gdwFlatSsSelectorBase;
	DWORD	gf32BitExecMode=FALSE;
	DWORD	gdwCsSelector;
	DWORD 	gdwUsedSegSelector;

// Set the exponent if the memory address is an MM register.
// gFpContext is defined below..
#define SetExponent(pMmReg)	\
	if ( ((BYTE *)pMmReg > (BYTE *)gFpContext) && ((BYTE *)pMmReg < (BYTE *)gFpContext+sizeof(FP_CONTEXT)) ) \
		((U80 *)pMmReg)->Exponent |= 0xffff;

#endif /* WIN_VXD */

/* NOT A WIN_VXD - This is the else clause */
#if !defined(WIN_VXD)
	#define MapAddressFlat(x,y,z)	TRUE
	#define SetExponent(pMmReg)
#endif 
/* -------- END WIN_VXD ----------------------------------------------*/
/* -------- END WIN_VXD ----------------------------------------------*/


/* -------- BEGIN WIN_APP For Win32 applications -----------------------*/
/* -------- BEGIN WIN_APP For Win32 applications -----------------------*/
#ifdef WIN_APP
	#include <windows.h>
	#include <windef.h>
	#include <winnt.h>
	#include <excpt.h>
	#include <winbase.h>

 	/* Defines ---------------------------------- */
	#define R__EAX		lucp->ContextRecord->Eax
	#define R__ECX		lucp->ContextRecord->Ecx
	#define R__EDX		lucp->ContextRecord->Edx
	#define R__EBX		lucp->ContextRecord->Ebx
	#define R__ESP		lucp->ContextRecord->Esp
	#define R__EBP		lucp->ContextRecord->Ebp
	#define R__ESI		lucp->ContextRecord->Esi
	#define R__EDI		lucp->ContextRecord->Edi
	#define R__EIP		lucp->ContextRecord->Eip


	/* Globals ---------------------------------*/
	extern PEXCEPTION_POINTERS	lucp;
	extern int	continue_search;

#endif /* WIN_APP */
/* -------- END   WIN_APP For Win32 applications -----------------------*/



/* -------- BEGIN UNIX_APP ---------------------------------------------*/
#if !defined(WIN_APP) && !defined(WIN_VXD)
	#define UNIX_APP
    #include <signal.h>
	#include <ucontext.h>
	
	/* Defines ----------------------------------*/
	#define R__EAX lucp->uc_mcontext.gregs[EAX]
	#define R__ECX lucp->uc_mcontext.gregs[ECX]
	#define R__EDX lucp->uc_mcontext.gregs[EDX]
	#define R__EBX lucp->uc_mcontext.gregs[EBX]
	#define R__ESP lucp->uc_mcontext.gregs[UESP]
	#define R__EBP lucp->uc_mcontext.gregs[EBP]
	#define R__ESI lucp->uc_mcontext.gregs[ESI]
	#define R__EDI lucp->uc_mcontext.gregs[EDI]
	#define R__EIP lucp->uc_mcontext.gregs[EIP]
	/* Globals ----------------------------------*/
	static	ucontext_t	ucp;
	static	ucontext_t	*lucp;


#endif	/* UNIX APP */



